@@ -10,6 +10,9 @@
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
+ * 
+ *==================================================================
+ * modified by fduncanh 2021-2023
  */
 
 #ifndef RAOP_RTP_H
@@ -29,8 +32,8 @@ typedef struct raop_rtp_s raop_rtp_t;
 raop_rtp_t *raop_rtp_init(logger_t *logger, raop_callbacks_t *callbacks, raop_ntp_t *ntp, const unsigned char *remote, 
                           int remotelen, const unsigned char *aeskey, const unsigned char *aesiv);
 
-void raop_rtp_start_audio(raop_rtp_t *raop_rtp, int use_udp, unsigned short control_rport,
-                          unsigned short *control_lport, unsigned short *data_lport, unsigned char ct);
+void raop_rtp_start_audio(raop_rtp_t *raop_rtp, int use_udp, unsigned short *control_rport, unsigned short *control_lport,
+                          unsigned short *data_lport, unsigned char *ct, unsigned int *sr);
 
 void raop_rtp_set_volume(raop_rtp_t *raop_rtp, float volume);
 void raop_rtp_set_metadata(raop_rtp_t *raop_rtp, const char *data, int datalen);
