@@ -1,6 +1,9 @@
 /**
  * RPiPlay - An open-source AirPlay mirroring server for Raspberry Pi
  * Copyright (C) 2019 Florian Draschbacher
+ * Modified extensively to become 
+ * UxPlay - An open-souce AirPlay mirroring server.
+ * Modifications Copyright (C) 2021-23 F. Duncanh
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,6 +28,9 @@
 #include <string>
 #include <vector>
 #include <fstream>
+#include <sstream>
+#include <iterator>
+#include <sys/stat.h>
 
 #ifdef _WIN32  /*modifications for Windows compilation */
 #include <glib.h>
@@ -36,6 +42,8 @@
 #include <sys/utsname.h>
 #include <sys/socket.h>
 #include <ifaddrs.h>
+#include <sys/types.h>
+#include <pwd.h>
 # ifdef __linux__
 # include <netpacket/packet.h>
 # else
@@ -51,8 +59,10 @@
 #include "renderers/video_renderer.h"
 #include "renderers/audio_renderer.h"
 
-#define VERSION "1.62"
+#define VERSION "1.65"
 
+#define SECOND_IN_USECS 1000000
+#define SECOND_IN_NSECS 1000000000UL
 #define DEFAULT_NAME "UxPlay"
 #define DEFAULT_DEBUG_LOG false
 #define LOWEST_ALLOWED_PORT 1024
@@ -64,7 +74,10 @@ static std::string server_name = DEFAULT_NAME;
 static dnssd_t *dnssd = NULL;
 static raop_t *raop = NULL;
 static logger_t *render_logger = NULL;
-
+static bool audio_sync = false;
+static bool video_sync = true;
+static int64_t audio_delay_alac = 0;
+static int64_t audio_delay_aac = 0;
 static bool relaunch_video = false;
 static bool reset_loop = false;
 static unsigned int open_connections= 0;
@@ -73,7 +86,7 @@ static videoflip_t videoflip[2] = { NONE , NONE };
 static bool use_video = true;
 static unsigned char compression_type = 0;
 static std::string audiosink = "autoaudiosink";
-static std::string audiodelay = "";
+static int  audiodelay = -1;
 static bool use_audio = true;
 static bool new_window_closing_behavior = true;
 static bool close_window;
@@ -107,7 +120,7 @@ static bool bt709_fix = false;
 static int max_connections = 2;
 static unsigned short raop_port;
 static unsigned short airplay_port;
-static uint64_t ntp_start_time, gst_start_time;
+static uint64_t remote_clock_offset = 0;
 
 /* 95 byte png file with a 1x1 white square (single pixel): placeholder for coverart*/
 static const unsigned char empty_image[] = {
@@ -234,7 +247,6 @@ guint g_unix_signal_add(gint signum, GSourceFunc handler, gpointer user_data) {
 #endif
 
 static void main_loop()  {
-    guint connection_watch_id = 0;
     guint gst_bus_watch_id = 0;
     GMainLoop *loop = g_main_loop_new(NULL,FALSE);
     relaunch_video = false;
@@ -255,12 +267,43 @@ static void main_loop()  {
 }    
 
 static int parse_hw_addr (std::string str, std::vector<char> &hw_addr) {
-    for (int i = 0; i < str.length(); i += 3) {
+    for (int i = 0; i < (int) str.length(); i += 3) {
         hw_addr.push_back((char) stol(str.substr(i), NULL, 16));
     }
     return 0;
 }
 
+static std::string find_uxplay_config_file() {
+    std::string no_config_file = "";
+    const char *homedir = NULL;
+    const char *uxplayrc = NULL;
+    std::string config0, config1, config2;
+    struct stat sb;
+    uxplayrc = getenv("UXPLAYRC");   /* first look for $UXPLAYRC */
+    if (uxplayrc) {
+        config0 = uxplayrc;
+        if (stat(config0.c_str(), &sb) == 0) return config0;
+    }
+    homedir = getenv("XDG_CONFIG_HOMEDIR");
+    if (homedir == NULL) {
+        homedir = getenv("HOME");
+    }
+#ifndef _WIN32
+    if (homedir == NULL){
+        homedir = getpwuid(getuid())->pw_dir;
+    }
+#endif
+    if (homedir) {
+      config1 = homedir;
+      config1.append("/.uxplayrc");
+      if (stat(config1.c_str(), &sb) == 0) return config1;  /* look for ~/.uxplayrc */
+      config2 = homedir;
+      config2.append("/.config/uxplayrc"); /* look for ~/.config/uxplayrc */
+      if (stat(config2.c_str(), &sb) == 0) return config2;
+    }
+    return no_config_file;
+}
+
 static std::string find_mac () {
 /*  finds the MAC address of a network interface *
  *  in a Windows, Linux, *BSD or macOS system.   */
@@ -288,7 +331,7 @@ static std::string find_mac () {
             }
             mac.erase();
             for (int i = 0; i < 6; i++) {
-                sprintf(str,"%02x", int(address->PhysicalAddress[i]));
+                snprintf(str, sizeof(str), "%02x", int(address->PhysicalAddress[i]));
                 mac = mac + str;
                 if (i < 5) mac = mac + ":";
             }
@@ -300,7 +343,7 @@ static std::string find_mac () {
 #else
     struct ifaddrs *ifap, *ifaptr;
     int non_null_octets = 0;
-    unsigned char octet[6], *ptr;
+    unsigned char octet[6];
     if (getifaddrs(&ifap) == 0) {
         for(ifaptr = ifap; ifaptr != NULL; ifaptr = ifaptr->ifa_next) {
             if(ifaptr->ifa_addr == NULL) continue;
@@ -312,7 +355,7 @@ static std::string find_mac () {
             }
 #else    /* macOS and *BSD */
             if (ifaptr->ifa_addr->sa_family != AF_LINK) continue;
-            ptr = (unsigned char *) LLADDR((struct sockaddr_dl *) ifaptr->ifa_addr);
+            unsigned char *ptr = (unsigned char *) LLADDR((struct sockaddr_dl *) ifaptr->ifa_addr);
             for (int i= 0; i < 6 ; i++) {
                 if ((octet[i] = *ptr) != 0) non_null_octets++;
                 ptr++;
@@ -321,7 +364,7 @@ static std::string find_mac () {
             if (non_null_octets) {
                 mac.erase();
                 for (int i = 0; i < 6 ; i++) {
-                    sprintf(str,"%02x", octet[i]);
+                    snprintf(str, sizeof(str), "%02x", octet[i]);
                     mac = mac + str;
                     if (i < 5) mac = mac + ":";
                 }
@@ -354,11 +397,17 @@ static std::string random_mac () {
 }
 
 static void print_info (char *name) {
-    printf("UxPlay %s: An open-source AirPlay mirroring server based on RPiPlay\n", VERSION);
-    printf("Usage: %s [-n name] [-s wxh] [-p [n]]\n", name);
+    printf("UxPlay %s: An open-source AirPlay mirroring server.\n", VERSION);
+    printf("=========== Website: https://github.com/FDH2/UxPlay ==========\n");
+    printf("Usage: %s [-n name] [-s wxh] [-p [n]] [(other options)]\n", name);
     printf("Options:\n");
     printf("-n name   Specify the network name of the AirPlay server\n");
-    printf("-nh       Do not add \"@hostname\" at the end of the AirPlay server name\n");
+    printf("-nh       Do not add \"@hostname\" at the end of AirPlay server name\n");
+    printf("-vsync [x]Mirror mode: sync audio to video using timestamps (default)\n");
+    printf("          x is optional audio delay: millisecs, decimal, can be neg.\n");
+    printf("-vsync no Switch off audio/(server)video timestamp synchronization \n");
+    printf("-async [x]Audio-Only mode: sync audio to client video (default: no)\n");
+    printf("-async no Switch off audio/(client)video timestamp synchronization\n");
     printf("-s wxh[@r]Set display resolution [refresh_rate] default 1920x1080[@60]\n");
     printf("-o        Set display \"overscanned\" mode on (not usually needed)\n");
     printf("-fs       Full-screen (only works with X11, Wayland and VAAPI)\n");
@@ -373,13 +422,13 @@ static void print_info (char *name) {
     printf("          nvdec, nvh264dec, vaapih64dec, vtdec,etc.\n");
     printf("          choices: avdec_h264,vaapih264dec,nvdec,nvh264dec,v4l2h264dec\n");
     printf("-vc ...   Choose the GStreamer videoconverter; default \"videoconvert\"\n");
-    printf("          another choice when using v4l2h264decode: v4l2convert\n");
+    printf("          another choice when using v4l2h264dec: v4l2convert\n");
     printf("-vs ...   Choose the GStreamer videosink; default \"autovideosink\"\n");
     printf("          some choices: ximagesink,xvimagesink,vaapisink,glimagesink,\n");
     printf("          gtksink,waylandsink,osximagesink,kmssink,d3d11videosink etc.\n");
     printf("-vs 0     Streamed audio only, with no video display window\n");
     printf("-v4l2     Use Video4Linux2 for GPU hardware h264 decoding\n");
-    printf("-bt709    A workaround (bt709 color) that may be needed with -rpi\n"); 
+    printf("-bt709    A workaround (bt709 color) sometimes needed on RPi\n"); 
     printf("-rpi      Same as \"-v4l2\" (for RPi=Raspberry Pi).\n");
     printf("-rpigl    Same as \"-rpi -vs glimagesink\" for RPi.\n");
     printf("-rpifb    Same as \"-rpi -vs kmssink\" for RPi using framebuffer.\n");
@@ -388,7 +437,7 @@ static void print_info (char *name) {
     printf("          some choices:pulsesink,alsasink,pipewiresink,jackaudiosink,\n");
     printf("          osssink,oss4sink,osxaudiosink,wasapisink,directsoundsink.\n");
     printf("-as 0     (or -a)  Turn audio off, streamed video only\n");
-    printf("-ao x.y   Audio offset time in seconds (default 0.0) in Audio-only mode.\n");
+    printf("-al x     Audio latency in seconds (default 0.25) reported to client.\n");
     printf("-ca <fn>  In Airplay Audio (ALAC) mode, write cover-art to file <fn>\n");
     printf("-reset n  Reset after 3n seconds client silence (default %d, 0=never)\n", NTP_TIMEOUT_LIMIT);
     printf("-nc       do Not Close video window when client stops mirroring\n");
@@ -409,6 +458,9 @@ static void print_info (char *name) {
     printf("-d        Enable debug logging\n");
     printf("-v        Displays version information\n");
     printf("-h        Displays this help\n");
+    printf("Startup options in $UXPLAYRC, ~/.uxplayrc, or ~/.config/uxplayrc are\n");
+    printf("applied first (command-line options may modify them): format is one \n");
+    printf("option per line, no initial \"-\"; lines starting with \"#\" are ignored.\n");
 }
 
 bool option_has_value(const int i, const int argc, std::string option, const char *next_arg) {
@@ -552,6 +604,46 @@ static void parse_arguments (int argc, char *argv[]) {
             server_name = std::string(argv[++i]);
         } else if (arg == "-nh") {
             do_append_hostname = false;
+        } else if (arg == "-async") {
+            audio_sync = true;
+	    if (i <  argc - 1) {
+                if (strlen(argv[i+1]) == 2 && strncmp(argv[i+1], "no", 2) == 0) {
+                    audio_sync = false;
+                    i++;
+                    continue;
+		}
+                char *end;
+                int n = (int) (strtof(argv[i + 1], &end) * 1000);
+                if (*end == '\0') {
+                    i++;
+                    if (n > -SECOND_IN_USECS && n < SECOND_IN_USECS) {
+                        audio_delay_alac = n * 1000; /* units are nsecs */
+                    } else {
+		      fprintf(stderr, "invalid -async %s: requested delays must be smaller than +/- 1000 millisecs\n", argv[i] );
+                        exit (1);
+                    }
+                }
+            }
+        } else if (arg == "-vsync") {
+            video_sync = true;
+	    if (i <  argc - 1) {
+                if (strlen(argv[i+1]) == 2 && strncmp(argv[i+1], "no", 2) == 0) {
+                    video_sync = false;
+                    i++;
+                    continue;
+                }
+                char *end;
+                int n = (int) (strtof(argv[i + 1], &end) * 1000);
+                if (*end == '\0') {
+                    i++;
+                    if (n > -SECOND_IN_USECS && n < SECOND_IN_USECS) {
+                        audio_delay_aac = n * 1000;     /* units are nsecs */
+                    } else {
+		      fprintf(stderr, "invalid -vsync %s: requested delays must be smaller than +/- 1000 millisecs\n", argv[i]);
+                        exit (1);
+                    }
+                }
+            }
         } else if (arg == "-s") {
             if (!option_has_value(i, argc, argv[i], argv[i+1])) exit(1);
             std::string value(argv[++i]);
@@ -607,7 +699,7 @@ static void parse_arguments (int argc, char *argv[]) {
             use_audio = false;
         } else if (arg == "-d") {
             debug_log = !debug_log;
-        } else if (arg == "-h") {
+        } else if (arg == "-h"  || arg == "--help" || arg == "-?" || arg == "-help") {
             print_info(argv[0]);
             exit(0);
         } else if (arg == "-v") {
@@ -738,27 +830,21 @@ static void parse_arguments (int argc, char *argv[]) {
             bt709_fix = true;
         } else if (arg == "-nohold") {
             max_connections = 3;
-        } else if (arg == "-ao") {
+        } else if (arg == "-al") {
 	    int n;
             char *end;
             if (i < argc - 1 && *argv[i+1] != '-') {
-                n = (int) (1000 * strtof(argv[++i], &end));
-                if (*end == '\0' && n >=0 && n <= 10000) {
-                    audiodelay.erase();
-                    if (n > 0) {
-                        char* delay = new char[6];
-                        snprintf(delay, 6, "%d", n);
-                        audiodelay = delay;
-                        delete[] delay;
-                    }
+	      n = (int) (strtof(argv[++i], &end) * SECOND_IN_USECS);
+                if (*end == '\0' && n >=0 && n <= 10 * SECOND_IN_USECS) {
+                    audiodelay = n;
                     continue;
                 }
             }
-            fprintf(stderr, "invalid argument -ao %s: must be a decimal time offset in seconds, range [0,10]\n"
-                    "(like 5 or 4.8, which will be converted to a whole number of milliseconds)\n", argv[i]);
+            fprintf(stderr, "invalid argument -al %s: must be a decimal time offset in seconds, range [0,10]\n"
+                    "(like 5 or 4.8, which will be converted to a whole number of microseconds)\n", argv[i]);
             exit(1);
 	} else {
-            fprintf(stderr, "unknown option %s, stopping\n",argv[i]);
+            fprintf(stderr, "unknown option %s, stopping (for help use option \"-h\")\n",argv[i]);
             exit(1);
         }
     }
@@ -954,6 +1040,7 @@ static int start_dnssd(std::vector<char> hw_addr, std::string name) {
     return 0;
 }
 
+
 // Server callbacks
 extern "C" void conn_init (void *cls) {
     open_connections++;
@@ -965,6 +1052,12 @@ extern "C" void conn_destroy (void *cls) {
     //video_renderer_update_background(-1);
     open_connections--;
     //LOGD("Open connections: %i", open_connections);
+    if (open_connections == 0) {
+        remote_clock_offset = 0;
+        if (use_audio) {
+            audio_renderer_stop();
+        }
+    }
 }
 
 extern "C" void conn_reset (void *cls, int timeouts, bool reset_video) {
@@ -991,8 +1084,16 @@ extern "C" void audio_process (void *cls, raop_ntp_t *ntp, audio_decode_struct *
         dump_audio_to_file(data->data, data->data_len, (data->data)[0] & 0xf0);
     }
     if (use_audio) {
-      uint64_t pts = ((data->ntp_time * 1000)  + gst_start_time) - ntp_start_time; 
-      audio_renderer_render_buffer(data->data, &(data->data_len), &(data->seqnum), &pts);
+        if (!remote_clock_offset) {
+            remote_clock_offset = data->ntp_time_local - data->ntp_time_remote;
+        }
+        data->ntp_time_remote = data->ntp_time_remote + remote_clock_offset;
+        if (data->ct == 2 && audio_delay_alac) {
+            data->ntp_time_remote = (uint64_t) ((int64_t) data->ntp_time_remote  + audio_delay_alac);
+        } else if (audio_delay_aac) {
+            data->ntp_time_remote = (uint64_t) ((int64_t) data->ntp_time_remote + audio_delay_aac);
+        }
+      audio_renderer_render_buffer(data->data, &(data->data_len), &(data->seqnum), &(data->ntp_time_remote));
     }
 }
 
@@ -1001,8 +1102,11 @@ extern "C" void video_process (void *cls, raop_ntp_t *ntp, h264_decode_struct *d
         dump_video_to_file(data->data, data->data_len);
     }
     if (use_video) {
-        uint64_t pts = ((data->ntp_time * 1000)  + gst_start_time) - ntp_start_time; 
-        video_renderer_render_buffer(data->data, &(data->data_len), &(data->nal_count), &pts);
+        if (!remote_clock_offset) {
+            remote_clock_offset = data->ntp_time_local - data->ntp_time_remote;
+        }
+        data->ntp_time_remote = data->ntp_time_remote + remote_clock_offset;
+        video_renderer_render_buffer(data->data, &(data->data_len), &(data->nal_count), &(data->ntp_time_remote));
     }
 }
 
@@ -1045,7 +1149,7 @@ extern "C" void audio_get_format (void *cls, unsigned char *ct, unsigned short *
     audio_type = type;
     
     if (use_audio) {
-        audio_renderer_start(ct);
+      audio_renderer_start(ct);
     }
 
     if (coverart_filename.length()) {
@@ -1130,7 +1234,6 @@ extern "C" void log_callback (void *cls, int level, const char *msg) {
 }
 
 int start_raop_server (unsigned short display[5], unsigned short tcp[3], unsigned short udp[3], bool debug_log) {
-    int dnssd_error;
     raop_callbacks_t raop_cbs;
     memset(&raop_cbs, 0, sizeof(raop_cbs));
     raop_cbs.conn_init = conn_init;
@@ -1165,6 +1268,7 @@ int start_raop_server (unsigned short display[5], unsigned short tcp[3], unsigne
  
     if (show_client_FPS_data) raop_set_plist(raop, "clientFPSdata", 1);
     raop_set_plist(raop, "max_ntp_timeouts", max_ntp_timeouts);
+    if (audiodelay >= 0) raop_set_plist(raop, "audio_delay_micros", audiodelay);
 
     /* network port selection (ports listed as "0" will be dynamically assigned) */
     raop_set_tcp_ports(raop, tcp);
@@ -1201,9 +1305,89 @@ static void stop_raop_server () {
     return;
 }
 
+static void read_config_file(const char * filename, const char * uxplay_name) {
+    std::string config_file = filename;
+    std::string option_char = "-";
+    std::vector<std::string> options;
+    options.push_back(uxplay_name);
+    std::ifstream file(config_file);
+    if (file.is_open()) {
+        fprintf(stdout,"UxPlay: reading configuration from  %s\n", config_file.c_str());
+        std::string line;
+        while (std::getline(file, line)) {
+            if (line[0] == '#') continue;
+            //  first process line into separate option items with '\0' as delimiter
+            bool is_part_of_item, in_quotes;
+            char endchar;
+            is_part_of_item = false;
+            for (int i = 0; i < (int) line.size(); i++) {
+                if (is_part_of_item == false) {
+                    if (line[i] == ' ') {
+                        line[i] = '\0';
+                    } else {
+                        // start of new item
+                        is_part_of_item = true;
+                        switch (line[i]) {
+                        case '\'':
+                        case '\"':
+                            endchar = line[i];
+                            line[i] = '\0';
+                            in_quotes = true;
+                            break;
+                        default:
+                            in_quotes = false;
+		            endchar = ' ';
+		            break;
+                        }
+                    }
+                } else {
+                    /* previous character was inside this item */
+                    if (line[i] == endchar) {
+                        if (in_quotes) {
+                            /* cases where endchar is inside quoted item */
+                            if (i > 0 && line[i - 1] == '\\') continue;
+                            if (i + 1 < (int) line.size() && line[i + 1] != ' ') continue;
+		        }
+                        line[i] =  '\0';
+                        is_part_of_item = false;
+                    }
+                }
+            }
+
+            // now tokenize the processed line   
+            std::istringstream iss(line);
+            std::string token;
+            bool first = true;
+            while (std::getline(iss, token, '\0')) {
+                if (token.size() > 0) {
+                    if (first) {
+                        options.push_back(option_char + token.c_str());
+                        first = false;
+                    } else {
+                        options.push_back(token.c_str());
+                    }
+		}
+	    }
+	}
+        file.close();
+    } else {
+        fprintf(stderr,"UxPlay: failed to open configuration file at %s\n", config_file.c_str());
+    }
+    if (options.size() > 1) {
+        int argc = options.size();
+        char **argv = (char **) malloc(sizeof(char*) * argc);
+        for (int i = 0; i < argc; i++) {
+            argv[i] = (char *) options[i].c_str();
+        }
+        parse_arguments (argc, argv);
+        free (argv);
+    }
+}
+
 int main (int argc, char *argv[]) {
     std::vector<char> server_hw_addr;
     std::string mac_address;
+    std::string config_file = "";
 
 #ifdef SUPPRESS_AVAHI_COMPAT_WARNING
     // suppress avahi_compat nag message.  avahi emits a "nag" warning (once)
@@ -1212,6 +1396,10 @@ int main (int argc, char *argv[]) {
     if (!getenv("AVAHI_COMPAT_NOWARN")) putenv(avahi_compat_nowarn);
 #endif
 
+    config_file = find_uxplay_config_file();
+    if (config_file.length()) {
+        read_config_file(config_file.c_str(), argv[0]);
+    }
     parse_arguments (argc, argv);
 
 #ifdef _WIN32    /*  use utf-8 terminal output; don't buffer stdout in WIN32 when debug_log = false */
@@ -1277,7 +1465,7 @@ int main (int argc, char *argv[]) {
         append_hostname(server_name);
     }
 
-    if (!gstreamer_init(&ntp_start_time, &gst_start_time)) {
+    if (!gstreamer_init()) {
         LOGE ("stopping");
         exit (1);
     }
@@ -1287,17 +1475,14 @@ int main (int argc, char *argv[]) {
     logger_set_level(render_logger, debug_log ? LOGGER_DEBUG : LOGGER_INFO);
 
     if (use_audio) {
-        if (audiodelay.c_str()[0]) {
-            LOGI("Audio-only ALAC streams will be delayed by %s milliseconds", audiodelay.c_str());
-        }
-        audio_renderer_init(render_logger, audiosink.c_str(), audiodelay.c_str());
+      audio_renderer_init(render_logger, audiosink.c_str(), &audio_sync, &video_sync);
     } else {
         LOGI("audio_disabled");
     }
 
     if (use_video) {
         video_renderer_init(render_logger, server_name.c_str(), videoflip, video_parser.c_str(),
-                            video_decoder.c_str(), video_converter.c_str(), videosink.c_str(), &fullscreen);
+                            video_decoder.c_str(), video_converter.c_str(), videosink.c_str(), &fullscreen, &video_sync);
         video_renderer_start();
     }
 
@@ -1350,7 +1535,8 @@ int main (int argc, char *argv[]) {
         if (use_video && close_window) {
             video_renderer_destroy();
             video_renderer_init(render_logger, server_name.c_str(), videoflip, video_parser.c_str(),
-                                video_decoder.c_str(), video_converter.c_str(), videosink.c_str(), &fullscreen);
+                                video_decoder.c_str(), video_converter.c_str(), videosink.c_str(), &fullscreen,
+                                &video_sync);
             video_renderer_start();
         }
         if (relaunch_video) {
